import java.net.Socket;
import java.io.*;
import java.util.*;

public class RobotController implements Runnable
{
  private Socket sock;
  private BufferedReader reader;
  private BufferedWriter writer;
  
  private int Cols;
  private int Rows;
  private char[][] board;
  private int[][] history;
  
  private int myID;
  private String myIDString;
  private int myCapacity;
  private int myMoney;
  private int initialMoney;
  private Random rand = new Random();
  
  private HashMap heldPackages = new HashMap();
  private HashMap allPackages = new HashMap();
  private int heldWeight = 0;
  private int col=0, row=0;
  private int currentMove = 1;
  
  public RobotController(String host, int port)
  {
    try
    {
      sock = new Socket(host, port);
      reader = new BufferedReader(new InputStreamReader(sock.getInputStream(), "UTF-8"));
      writer = new BufferedWriter(new OutputStreamWriter(sock.getOutputStream(), "UTF-8"));
      new Thread(this).start();
    }
    catch(Exception e)
    {
      e.printStackTrace();
      shutdown();
    }
  }
  
  private void shutdown()
  {
    try
    {
      if(reader!=null)
        reader.close();
      if(writer!=null)
        writer.close();
      if(sock!=null)
        sock.close();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  }
  
  private void sendCommand(String command) throws IOException
  {
    writer.write(command+"\n");
    writer.flush();
    //System.out.println("C: "+command);
  }
  
  private String readLine() throws IOException
  {
    String line;
    while((line = reader.readLine())==null)
    {
      try
      {
        Thread.sleep(1000);
      }
      catch(InterruptedException ie)
      {
      }
    }
    //System.out.println("S: "+line);
    return line;
  }
  
  public void run()
  {
    try
    {
      //Initialize
      sendCommand("Player");
      //Get board
      StringTokenizer st = new StringTokenizer(readLine());
      Cols = Integer.parseInt(st.nextToken());
      Rows = Integer.parseInt(st.nextToken());
      board = new char[Cols][Rows];
      history = new int[Cols][Rows];
      for(int row=0; row<Rows; row++)
      {
        String s = readLine();
        for(int col=0; col<Cols; col++)
          board[col][row] = s.charAt(col);
      }
      //Get player configuration
      st = new StringTokenizer(readLine());
      myID = Integer.parseInt(st.nextToken());
      myIDString = "#"+myID;
      myCapacity = Integer.parseInt(st.nextToken());
      initialMoney = myMoney = Integer.parseInt(st.nextToken());
      //Get initial server response (with the initial location of all players)
      
      //Game play
      while(true)
      {
        //get server response (what happened to all robots)
        String response = readLine();
        int index = response.indexOf(myIDString);
        if(index>=0)
        {
          response = response.substring(index+myIDString.length());
          index = response.indexOf("#");
          if(index>=0)
          {
            response = response.substring(0,index);
          }
          st = new StringTokenizer(response);
          while(st.hasMoreTokens())
          {
            String token = st.nextToken();
            if(token.equals("N"))
            {
              row++;
            }
            else if(token.equals("S"))
            {
              row--;
            }
            else if(token.equals("E"))
            {
              col++;
            }
            else if(token.equals("W"))
            {
              col--;
            }
            else if(token.equals("X"))
            {
              col = Integer.parseInt(st.nextToken())-1;
            }
            else if(token.equals("Y"))
            {
              row = Integer.parseInt(st.nextToken())-1;
            }
            else if(token.equals("P"))
            {
              Package p = (Package)allPackages.get(new Integer(st.nextToken()));
              if(p!=null)
              {
                heldPackages.put(new Integer(p.id), p);
                heldWeight += p.weight;
              }
            }
            else if(token.equals("D"))
            {
              Integer key = new Integer(st.nextToken());
              if(heldPackages.containsKey(key))
              {
                Package p = (Package)heldPackages.get(key);
                heldPackages.remove(key);
                heldWeight -= p.weight;
              }
            }
          }
        }
        
        //get package list for this location
        String packages = readLine();
        st = new StringTokenizer(packages);
        Package[] packagesHere = null;
        int numPackagesHere = st.countTokens()/4;
        if(numPackagesHere>0)
          packagesHere = new Package[numPackagesHere];
        for(int i=0; i<numPackagesHere; i++)
        {
          Package p = new Package();
          p.id = Integer.parseInt(st.nextToken());
          p.destCol = Integer.parseInt(st.nextToken())-1;
          p.destRow = Integer.parseInt(st.nextToken())-1;
          p.weight = Integer.parseInt(st.nextToken());
          allPackages.put(new Integer(p.id), p);
          packagesHere[i] = p;
        }
        
        //determine and send move
        history[col][row] = currentMove++;
        //am I holding any packages that belong here?
        Iterator iterator = heldPackages.values().iterator();
        ArrayList toDrop = new ArrayList();
        while(iterator.hasNext())
        {
          Package p = (Package)iterator.next();
          if(p.destCol==col && p.destRow==row)
            toDrop.add(new Integer(p.id));
        }
        if(toDrop.size()>0)
        {
          String commandString = "1 Drop";
          for(int i=0; i<toDrop.size(); i++)
            commandString += " " + ((Integer)toDrop.get(i)).intValue();
          sendCommand(commandString);
          continue;
        }
        //are there packages here I could pick up?
        if(numPackagesHere>0)
        {
          int freeCapacity = myCapacity-heldWeight;
          bubbleSortByPackageWeight(packagesHere);
          ArrayList toPickUp = new ArrayList();
          for(int i=0; i<packagesHere.length&&packagesHere[i].weight<=freeCapacity; i++)
          {
            toPickUp.add(new Integer(packagesHere[i].id));
            freeCapacity-=packagesHere[i].weight;
          }
          if(toPickUp.size()>0)
          {
            String commandString = "1 Pick";
            for(int i=0; i<toPickUp.size(); i++)
              commandString += " " + ((Integer)toPickUp.get(i)).intValue();
            sendCommand(commandString);
            continue;
          }
        }
        //get possible moves
        boolean north = row+1<Rows && (board[col][row+1]=='.' || board[col][row+1]=='@');
        boolean south = row-1>=0 && (board[col][row-1]=='.' || board[col][row-1]=='@');
        boolean east = col+1<Cols && (board[col+1][row]=='.' || board[col+1][row]=='@');
        boolean west = col-1>=0 && (board[col-1][row]=='.' || board[col-1][row]=='@');
        //do random weighted move
        String direction = "";
        int best = Integer.MAX_VALUE;
        int hist;
        if(north && (hist=history[col][row+1])<=best)
        {
          if(hist==best)
            direction+="N";
          else
            direction = "N";
          best = hist;
        }
        if(south && (hist=history[col][row-1])<=best)
        {
          if(hist==best)
            direction+="S";
          else
            direction = "S";
          best = hist;
        }
        if(east && (hist=history[col+1][row])<=best)
        {
          if(hist==best)
            direction+="E";
          else
            direction = "E";
          best = hist;
        }
        if(west && (hist=history[col-1][row])<=best)
        {
          if(hist==best)
            direction+="W";
          else
            direction = "W";
          best = hist;
        }
        sendCommand("1 Move " + direction.charAt(rand.nextInt(direction.length())));
       }
      
    }
    catch(Exception e)
    {
      //e.printStackTrace();
      shutdown();
    }
  }
  
  public static void bubbleSortByPackageWeight(Package[] A)
  {
    for(int i=0; i<A.length; i++)
      for(int j=A.length-1; j>i; j--)
        if(A[j].weight < A[j-1].weight)
        {
          Package temp = A[j];
          A[j] = A[j-1];
          A[j-1] = temp;
        }
  }
  
  public class Package
  {
    public int weight;
    public int destCol;
    public int destRow;
    public int id;
  }
}

